#include "ego_interface.h"
#include <iomanip>
#include <iostream>

namespace Localization {
namespace Ego {

EgoInterface::EgoInterface()
{
    std::cout<<"Ego Interface constructed"<<std::endl;
}

void EgoInterface::Init() {
  if (isUsePredict_) {
    predictThreadFlag_ = true;
    predictThread_= std::thread([this] {PredictThread();});
  }
  updateThreadFlag_ = true;
  updateThread_ = std::thread([this] {UpdateThread();});
  isKFInit_ = false;
  std::cout<<"Ego Interface Inited"<<std::endl;
}

void EgoInterface::StopThread(){
  predictThreadFlag_ = false;
  updateThreadFlag_ = false;
}

EgoInterface::~EgoInterface()
{
  isKFInit_ = false;
  StopThread();
  usleep(5e5);
}

bool EgoInterface::InitKF(std::shared_ptr<BaseMeasure> &ptrM)
{
  kf_.InitFilterState(ptrM);
  return kf_.GetInitFlag();
}

void EgoInterface::PredictThread()
{
  while (predictThreadFlag_) {
    usleep(1e4);
    State tempState;
    kf_.GetState(tempState);
    auto aimTime = tempState.time + bufferLenth;
    if (isKFInit_) {
      std::vector<std::shared_ptr<BaseMeasure>> tempBuf;
      {
        std::unique_lock<std::mutex> lock(mutexMeasureBuffer_);
        tempBuf = measureBuffer_;
      }
      if ((tempState.time < aimTime) && (aimTime < tempState.time + 1)) {
        kf_.PredictToTime(aimTime, tempState);
      }
    }
    LocalizationResult kfOutput;
    if (isKFInit_) {
      SetLocalizationResult(tempState,kfOutput);
      SetEgoOutput(kfOutput);
    }
  }
}

void EgoInterface::UpdateThread()
{
  while (updateThreadFlag_) {
    usleep(1e4);
    bool getMeaSucc = true;
    std::vector<std::shared_ptr<BaseMeasure>> tempBuf;
    {
      std::unique_lock<std::mutex> lock(mutexMeasureBuffer_);
      tempBuf = measureBuffer_;
    }
    while (getMeaSucc) {
      std::shared_ptr<BaseMeasure> ptrM;
      GetOneMeaFromBuf(tempBuf,ptrM);
      if (ptrM == nullptr) {
        getMeaSucc = false;
        continue;
      }
      if (isKFInit_ == false) {
        if (InitKF(ptrM)) {
          isKFInit_ = true;
          State initedState;
          kf_.GetState(initedState);
          std::cout<<"初始化成功,state是"<<std::endl;
          initedState.Print();
        }
        usleep(1e3);
        continue;
      }
      kf_.Compute(ptrM);
      if (kf_.GetComputeFlag() == false) {
        isKFInit_ = false;
        getMeaSucc = false;
        ResetMeaBuf();
        std::cerr<<"UpdateThread():ego is going to be reset,meaBuf Cleaned"<<std::endl;
        continue;
      }
      if (isUsePredict_ == false) {
        LocalizationResult kfOutput;
        State outputState;
        kf_.GetState(outputState);
        if (isKFInit_) {
          SetLocalizationResult(outputState,kfOutput);
          SetEgoOutput(kfOutput);
        }
      }
    }
  }
}

void EgoInterface::AddMeasureToBuffer(std::shared_ptr<Ego::BaseMeasure> &pMeasure)
{
  {
    std::unique_lock<std::mutex> lock(mutexMeasureBuffer_);
    if (measureBuffer_.empty()) {
      measureBuffer_.emplace_back(pMeasure);
      return;
    }
    if (measureBuffer_.size() > 600U) {
      measureBuffer_.assign(measureBuffer_.begin() + 300, measureBuffer_.end());
    }

    static int lastMeasureType = 0;
    if (pMeasure->stamp_ == measureBuffer_.back()->stamp_) {
      if (lastMeasureType != static_cast<int>(pMeasure->type_)) {
        lastMeasureType = static_cast<int>(pMeasure->type_);
        measureBuffer_.emplace_back(pMeasure);
      }
      return;
    }
    lastMeasureType = static_cast<int>(pMeasure->type_);
    if (pMeasure->stamp_ > measureBuffer_.back()->stamp_) {
      measureBuffer_.emplace_back(pMeasure);
      return;
    }
    for(int i = 0; i < measureBuffer_.size(); i++) {
      if (pMeasure->stamp_ < measureBuffer_[i]->stamp_) {
        measureBuffer_.insert(measureBuffer_.begin() + i, pMeasure);
        break;
      }
    }
  } // mutex
}

void EgoInterface::GetOneMeaFromBuf(std::vector<std::shared_ptr<BaseMeasure>> &meaBuf, std::shared_ptr<Ego::BaseMeasure> &ptrM)
{
  if (meaBuf.empty()) {
//    std::cout<<"GetOneMeaFromBuf(): meaBuf is empty, in case of null ptr,return"<<std::endl;
    return;
  }
  if (meaBuf.back()->stamp_ >= kf_.GetStateTime() + bufferLenth) {
    for (size_t i = 0U; i < meaBuf.size(); i++) {
      if (meaBuf.back()->stamp_ - meaBuf.front()->stamp_ < bufferLenth) {
        break;
      }
      if (meaBuf[i]->stamp_ >= kf_.GetStateTime()) {
        if (meaBuf[i]->isKFUsed_ == true) {
          continue;
        }
        meaBuf[i]->isKFUsed_ = true;
        ptrM = meaBuf[i];
        meaBuf.assign(meaBuf.begin() + static_cast<long>(i) + 1, meaBuf.end());
        break;
      }
    }
  }
  if (!meaBuf.empty()) {
    double delat = meaBuf.back()->stamp_ - kf_.GetStateTime();
    if (delat > 3.0) {
      std::cout<<"GetOneMeaFromBuf():meaBuf.back()比State的时间大太多，观测可能阻塞，delta为："<<delat<<"秒"<<std::endl;
    }
  }
}

void EgoInterface::ResetMeaBuf()
{
  std::unique_lock<std::mutex> lock(mutexMeasureBuffer_);
  measureBuffer_.clear();
}

void EgoInterface::AddMeasureToRecBuf(std::shared_ptr<BaseMeasure> &pMeasure)
{
  std::unique_lock<std::mutex> lock(mutexRecLocBuffer_);
  if (recLocBuffer_.empty()) {
    recLocBuffer_.emplace_back(pMeasure);
    return;
  }
  if (recLocBuffer_.size() > 600U) {
    recLocBuffer_.assign(recLocBuffer_.begin() + 300, recLocBuffer_.end());
  }
  if (pMeasure->stamp_ > measureBuffer_.back()->stamp_) {
    measureBuffer_.emplace_back(pMeasure);
    return;
  }
}

void EgoInterface::SetRtkInput(InsInfo &rtk)
{
  static double lastStamp = rtk.stamp_;
  double curStamp = rtk.stamp_;
  if (curStamp - lastStamp > 1e-3) {
    std::shared_ptr<Localization::Ego::BaseMeasure> pRtk = std::make_shared<Localization::Ego::RtkMeasure>(
        Localization::Ego::MeasurementType::RTK, rtk);
    AddMeasureToBuffer(pRtk);
  }
  lastStamp = curStamp;
}

void EgoInterface::SetImuInput(ImuInfo &imu)
{
  static double lastStamp = imu.stamp_;
  double curStamp = imu.stamp_;
  if (curStamp - lastStamp > 1e-3) {
    std::shared_ptr<Localization::Ego::BaseMeasure> pImu = std::make_shared<Localization::Ego::ImuMeasure>(
        Localization::Ego::MeasurementType::IMU, imu);
    AddMeasureToBuffer(pImu);
  }
  lastStamp = curStamp;
}

void EgoInterface::SetVehInput(VehInfo &veh)
{
  static double lastStamp = veh.stamp_;
  double curStamp = veh.stamp_;
  if (curStamp - lastStamp > 1e-3) {
    std::shared_ptr<Localization::Ego::BaseMeasure> pWss = std::make_shared<Localization::Ego::VehMeasure>(
        Localization::Ego::MeasurementType::WHEEL, veh);
    AddMeasureToBuffer(pWss);
  }
  lastStamp = curStamp;
}

void EgoInterface::SetVLocInput(RecLocInfo &vLoc)
{
  static double lastStamp = vLoc.stamp_;
  double curStamp = vLoc.stamp_;
  if (curStamp - lastStamp > 1e-3) {
    std::shared_ptr<Localization::Ego::BaseMeasure> pVloc = std::make_shared<Localization::Ego::VLocMeasure>(
            Localization::Ego::MeasurementType::VSL, vLoc);
      AddMeasureToBuffer(pVloc);
  }
  lastStamp = curStamp;
}

void EgoInterface::SetCheckMeaInput(CheckMeaInfo &checkMea)
{
  static double lastStamp = checkMea.stamp_;
  double curStamp = checkMea.stamp_;
  if (curStamp - lastStamp > 1e-4) {
    std::shared_ptr<Localization::Ego::BaseMeasure> pcheckMea = std::make_shared<Localization::Ego::CheckMeasure>(
      Localization::Ego::MeasurementType::CHECKMEASURE, checkMea);
    AddMeasureToBuffer(pcheckMea);
  }
  lastStamp = curStamp;
}

void EgoInterface::SetLocalizationResult(State &stateRes,LocalizationResult &output)
{
  uint8_t status = 0U;
  if (stateRes.isImuValid_ == true)
    status |= 1<<0;
  if (stateRes.isVehValid_ == true)
    status |= 1<<1;
  if (stateRes.isRtkValid_ == true)
    status |= 1<<2;
  if (stateRes.isVlocValid_ == true)
    status |= 1<<3;
  output.integratedState = status;
  output.stamp = stateRes.time;
  Eigen::Vector3d mapOffset_{0,0,0};
  output.posUTM = stateRes.p + mapOffset_;
  output.quat = stateRes.q;
  Eigen::Vector3d YPR = RotYPR2EulerYPR(stateRes.q.toRotationMatrix());
  output.attitude.x() = YPR.z();
  output.attitude.y() = YPR.y();
  output.attitude.z() = YPR.x();
  output.veloENU = stateRes.v;
  output.angularVelocity = stateRes.av;
  output.linearAcceleration = stateRes.acc;
  output.cov = stateRes.mP;
  output.accBias = stateRes.ba;
  output.gyroBias = stateRes.bg;
}

void EgoInterface::SetEgoOutput(LocalizationResult &kfRes)
{
  std::unique_lock<std::mutex> lock(mutexOutput_);
  egoOuput_ = kfRes;
}

void EgoInterface::GetEgoOutput(LocalizationResult &egoOuput)
{
  std::unique_lock<std::mutex> lock(mutexOutput_);
  egoOuput = egoOuput_;
}

} // namespace Ego
} // namespace Localizaiton

